@@ -19,8 +19,6 @@
 
 Mail::SpamAssassin::Message::Node - decode, render, and make available MIME message parts
 
-=head1 SYNOPSIS
-
 =head1 DESCRIPTION
 
 This module will encapsulate an email message and allow access to
@@ -186,7 +184,7 @@ sub header {
     $dec_value =~ s/\n[ \t]+/ /gs;
     $dec_value =~ s/\s+$//s;
     $dec_value =~ s/^\s+//s;
-    push @{ $self->{'headers'}->{$key} }, $self->_decode_header($dec_value,$key);
+    push @{ $self->{'headers'}->{$key} }, _decode_header($dec_value,$key);
 
     push @{ $self->{'raw_headers'}->{$key} }, $raw_value;
 
@@ -379,6 +377,55 @@ sub decode {
   }
 }
 
+# Detect endianness of UTF-16 encoded data
+sub detect_utf16 {
+	my $data = $_[0];  # could not avoid copying large strings
+	my $utf16le_clues = 0;
+	my $utf16be_clues = 0;
+	my $sum_h_e = 0;
+	my $sum_h_o = 0;
+	my $sum_l_e = 0;
+	my $sum_l_o = 0;
+	my $decoder = undef;
+
+	my @msg_h = unpack 'H' x length( $data ), $data;
+	my @msg_l = unpack 'h' x length( $data ), $data;
+
+	for( my $i = 0; $i < length( $data ); $i+=2 ) {
+		my $check_char = sprintf( "%01X%01X %01X%01X", hex $msg_h[$i], hex $msg_l[$i], hex $msg_h[$i+1], hex $msg_l[$i+1] );
+		$sum_h_e += hex $msg_h[$i];
+		$sum_h_o += hex $msg_h[$i+1];
+		$sum_l_e += hex $msg_l[$i];
+		$sum_l_o += hex $msg_l[$i+1];
+		if( $check_char =~ /20 00/ ) {
+			# UTF-16LE space char detected
+			$utf16le_clues++;
+		}
+		if( $check_char =~ /00 20/ ) {
+			# UTF-16BE space char detected
+			$utf16be_clues++;
+		}
+	}
+
+	# If we have 4x as many non-null characters in the odd bytes, we're probably UTF-16LE
+	$utf16le_clues++ if( ($sum_h_e + $sum_l_e) > ($sum_h_o + $sum_l_o)*4 );
+
+	# If we have 4x as many non-null characters in the even bytes, we're probably UTF-16BE
+	$utf16be_clues++ if( ($sum_h_o + $sum_l_o)*4 > ($sum_h_e + $sum_l_e) );
+
+	if( $utf16le_clues > $utf16be_clues ) {
+		dbg( "message: detect_utf16: UTF-16LE" );
+		$decoder = Encode::find_encoding("UTF-16LE");
+	} elsif( $utf16le_clues > $utf16be_clues ) {
+		dbg( "message: detect_utf16: UTF-16BE" );
+		$decoder = Encode::find_encoding("UTF-16BE");
+	} else {
+		dbg( "message: detect_utf16: Could not detect UTF-16 endianness" );
+	}
+
+	return $decoder;
+}
+
 # Look at a text scalar and determine whether it should be rendered
 # as text/html.
 #
@@ -400,25 +447,22 @@ sub _html_render {
 # then encode into UTF-8 octets if requested.
 #
 sub _normalize {
-  my $self = $_[0];
-# my $data = $_[1];  # avoid copying large strings
-  my $charset_declared = $_[2];
-  my $return_decoded = $_[3];  # true: Unicode characters, false: UTF-8 octets
+# my $data = $_[0];  # avoid copying large strings
+  my $charset_declared = $_[1];
+  my $return_decoded = $_[2];  # true: Unicode characters, false: UTF-8 octets
 
-  return $_[1]  unless $self->{normalize} && $enc_utf8;
-
-  warn "message: _normalize() was given characters, expected bytes: $_[1]\n"
-    if utf8::is_utf8($_[1]);
+  warn "message: _normalize() was given characters, expected bytes: $_[0]\n"
+    if utf8::is_utf8($_[0]);
 
   # workaround for Encode::decode taint laundering bug [rt.cpan.org #84879]
-  my $data_taint = substr($_[1], 0, 0);  # empty string, tainted like $data
+  my $data_taint = substr($_[0], 0, 0);  # empty string, tainted like $data
 
   if (!defined $charset_declared || $charset_declared eq '') {
     $charset_declared = 'us-ascii';
   }
 
   # number of characters with code above 127
-  my $cnt_8bits = $_[1] =~ tr/\x00-\x7F//c;
+  my $cnt_8bits = $_[0] =~ tr/\x00-\x7F//c;
 
   if (!$cnt_8bits &&
       $charset_declared =~
@@ -426,7 +470,7 @@ sub _normalize {
               ISO646-US )\z/xsi)
   { # declared as US-ASCII (a.k.a. ANSI X3.4-1986) and it really is
     dbg("message: kept, charset is US-ASCII as declared");
-    return $_[1];  # is all-ASCII, no need for decoding
+    return $_[0];  # is all-ASCII, no need for decoding
   }
 
   if (!$cnt_8bits &&
@@ -436,7 +480,7 @@ sub _normalize {
               Big5 | GBK | GB[ -]?18030 (?:-20\d\d)? )\z/xsi)
   { # declared as extended ASCII, but it is actually a plain 7-bit US-ASCII
     dbg("message: kept, charset is US-ASCII, declared %s", $charset_declared);
-    return $_[1];  # is all-ASCII, no need for decoding
+    return $_[0];  # is all-ASCII, no need for decoding
   }
 
   # Try first to strictly decode based on a declared character set.
@@ -444,19 +488,44 @@ sub _normalize {
   my $rv;
   if ($charset_declared =~ /^UTF-?8\z/i) {
     # attempt decoding as strict UTF-8  (flags: FB_CROAK | LEAVE_SRC)
-    if (eval { $rv = $enc_utf8->decode($_[1], 1|8); defined $rv }) {
+    if (eval { $rv = $enc_utf8->decode($_[0], 1|8); defined $rv }) {
       dbg("message: decoded as declared charset UTF-8");
-      return $_[1]  if !$return_decoded;
+      return $_[0]  if !$return_decoded;
       $rv .= $data_taint;  # carry taintedness over, avoid Encode bug
       return $rv;  # decoded
     } else {
-      dbg("message: failed decoding as declared charset UTF-8");
+      my $err = '';
+      if ($@) {
+        $err = $@; $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/;
+        $err = " ($err)";
+      }
+      dbg("message: failed decoding as declared charset UTF-8 ($err)");
+    }
+
+  } elsif ($charset_declared =~ /^UTF[ -]?16/i) {
+    # Handle cases where spammers use UTF-16 encoding without including a BOM
+    # or declaring endianness as reported at:
+    # https://bz.apache.org/SpamAssassin/show_bug.cgi?id=7252
+
+    my $decoder = detect_utf16( $_[0] );
+    if (eval { $rv = $decoder->decode($_[0], 1|8); defined $rv }) {
+      dbg("message: declared charset %s decoded as charset %s", $charset_declared, $decoder->name);
+      return $_[0]  if !$return_decoded;
+      $rv .= $data_taint;  # carry taintedness over, avoid Encode bug
+      return $rv;  # decoded
+    } else {
+      my $err = '';
+      if ($@) {
+        $err = $@; $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/;
+        $err = " ($err)";
+      }
+      dbg("message: failed decoding as declared charset %s%s", $charset_declared, $err);
     };
 
   } elsif ($cnt_8bits &&
-           eval { $rv = $enc_utf8->decode($_[1], 1|8); defined $rv }) {
+           eval { $rv = $enc_utf8->decode($_[0], 1|8); defined $rv }) {
     dbg("message: decoded as charset UTF-8, declared %s", $charset_declared);
-    return $_[1]  if !$return_decoded;
+    return $_[0]  if !$return_decoded;
     $rv .= $data_taint;  # carry taintedness over, avoid Encode bug
     return $rv;  # decoded
 
@@ -496,14 +565,19 @@ sub _normalize {
       dbg("message: failed decoding, no decoder for a declared charset %s",
           $chset);
     } else {
-      eval { $rv = $decoder->decode($_[1], 1|8) };  # FB_CROAK | LEAVE_SRC
+      my $err = '';
+      eval { $rv = $decoder->decode($_[0], 1|8) };  # FB_CROAK | LEAVE_SRC
+      if ($@) {
+        $err = $@; $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/;
+        $err = " ($err)";
+      }
       if (lc $chset eq lc $charset_declared) {
-        dbg("message: %s as declared charset %s",
-            defined $rv ? 'decoded' : 'failed decoding', $charset_declared);
+        dbg("message: %s as declared charset %s%s",
+            defined $rv ? 'decoded' : 'failed decoding', $charset_declared, $err);
       } else {
-        dbg("message: %s as charset %s, declared %s",
+        dbg("message: %s as charset %s, declared %s%s",
             defined $rv ? 'decoded' : 'failed decoding',
-            $chset, $charset_declared);
+            $chset, $charset_declared, $err);
       }
     }
   }
@@ -517,19 +591,27 @@ sub _normalize {
   if (!defined $rv && !$cnt_8bits) {
     dbg("message: kept, guessed charset is US-ASCII, declared %s",
         $charset_declared);
-    return $_[1];  # is all-ASCII, no need for decoding
+    return $_[0];  # is all-ASCII, no need for decoding
 
   } elsif (!defined $rv && $enc_w1252 &&
-      #             ASCII  NBSP (c) SHY  '   "  ...   '".-   TM
-      $_[1] !~ tr/\x00-\x7F\xA0\xA9\xAD\x82\x84\x85\x91-\x97\x99//c)
+     #             ASCII  NBSP (c) SHY  '   "  ...   '".-   TM
+     #$_[0] !~ tr/\x00-\x7F\xA0\xA9\xAD\x82\x84\x85\x91-\x97\x99//c)
+     # Bug 7656: Include latin1 diacritic letters to Windows-1252 autodetection,
+     # Encode::Detect::Detector might identify them as Windows-1255 (Hebrew!)
+      $_[0] !~ tr/\x00-\x7f\xa0\xa9\xad\x82\x84\x85\x91-\x97\x99\xc0-\xd6\xd8-\xde\xe0-\xf6\xf8-\xfe//c)
   { # ASCII + NBSP + SHY + some punctuation characters
     # NBSP (A0) and SHY (AD) are at the same position in ISO-8859-* too
     # consider also: AE (r), 80 Euro
-    eval { $rv = $enc_w1252->decode($_[1], 1|8) };  # FB_CROAK | LEAVE_SRC
+    my $err = '';
+    eval { $rv = $enc_w1252->decode($_[0], 1|8) };  # FB_CROAK | LEAVE_SRC
+    if ($@) {
+      $err = $@; $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/;
+      $err = " ($err)";
+    }
     # the above can't fail, but keep code general just in case
-    dbg("message: %s as guessed charset %s, declared %s",
+    dbg("message: %s as guessed charset %s, declared %s%s",
         defined $rv ? 'decoded' : 'failed decoding',
-        'Windows-1252', $charset_declared);
+        'Windows-1252', $charset_declared, $err);
   }
 
   # If we were unsuccessful so far, try some guesswork
@@ -541,7 +623,7 @@ sub _normalize {
     dbg("message: Encode::Detect::Detector not available, declared %s failed",
         $charset_declared);
   } else {
-    my $charset_detected = Encode::Detect::Detector::detect($_[1]);
+    my $charset_detected = Encode::Detect::Detector::detect($_[0]);
     if ($charset_detected && lc $charset_detected ne lc $charset_declared) {
       my $decoder = Encode::find_encoding($charset_detected);
       if (!$decoder && $charset_detected =~ /^GB[ -]?18030(?:-20\d\d)?\z/i) {
@@ -553,24 +635,34 @@ sub _normalize {
         dbg("message: failed decoding, no decoder for a detected charset %s",
             $charset_detected);
       } else {
-        eval { $rv = $decoder->decode($_[1], 1|8) };  # FB_CROAK | LEAVE_SRC
-        dbg("message: %s as detected charset %s, declared %s",
+        my $err = '';
+        eval { $rv = $decoder->decode($_[0], 1|8) };  # FB_CROAK | LEAVE_SRC
+        if ($@) {
+          $err = $@; $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/;
+          $err = " ($err)";
+        }
+        dbg("message: %s as detected charset %s, declared %s%s",
             defined $rv ? 'decoded' : 'failed decoding',
-            $charset_detected, $charset_declared);
+            $charset_detected, $charset_declared, $err);
       }
     }
   }
 
   if (!defined $rv) {  # all decoding attempts failed so far, probably garbage
     # go for Windows-1252 which can't fail
-    eval { $rv = $enc_w1252->decode($_[1]) };
-    dbg("message: %s as last-resort charset %s, declared %s",
+    my $err = '';
+    eval { $rv = $enc_w1252->decode($_[0]) };
+    if ($@) {
+      $err = $@; $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/;
+      $err = " ($err)";
+    }
+    dbg("message: %s as last-resort charset %s, declared %s%s",
         defined $rv ? 'decoded' : 'failed decoding',
-        'Windows-1252', $charset_declared);
+        'Windows-1252', $charset_declared, $err);
   }
 
   if (!defined $rv) {  # just in case - all decoding attempts failed so far
-    return $_[1];  # garbage-in / garbage-out, return unchanged octets
+    return $_[0];  # garbage-in / garbage-out, return unchanged octets
   }
   # decoding octets to characters was successful
   if (!$return_decoded) {
@@ -604,11 +696,8 @@ sub rendered {
     my $text = $self->decode;  # QP and Base64 decoding, bytes
     my $text_len = length($text);  # num of bytes in original charset encoding
 
-    # render text/html always, or any other text|text/plain part as text/html
-    # based on a heuristic which simulates a certain common mail client
-    if ($text ne '' && ($self->{'type'} =~ m{^text/html$}i ||
-		        ($self->{'type'} =~ m{^text/plain$}i &&
-		         _html_render(substr($text, 0, 23)))))
+    # render text/html always
+    if ($text ne '' && $self->{'type'} =~ m{^text/html$}i)
     {
       $self->{rendered_type} = 'text/html';
 
@@ -618,11 +707,12 @@ sub rendered {
         # Provide input to HTML::Parser as Unicode characters
         # which avoids a HTML::Parser bug in utf8_mode
         #   https://rt.cpan.org/Public/Bug/Display.html?id=99755
+        #   Note: the above bug was fixed in HTML-Parser 3.72, January 2016.
         # Avoid unnecessary step of encoding-then-decoding by telling
         # subroutine _normalize() to return Unicode text.  See Bug 7133
         #
         $character_semantics = 1;  # $text will be in characters
-        $text = $self->_normalize($text, $self->{charset}, 1); # bytes to chars
+        $text = _normalize($text, $self->{charset}, 1); # bytes to chars
       } elsif (!defined $self->{charset} ||
                $self->{charset} =~ /^(?:US-ASCII|UTF-8)\z/i) {
         # With some luck input can be interpreted as UTF-8, do not warn.
@@ -657,7 +747,7 @@ sub rendered {
     else {  # plain text
       if ($self->{normalize} && $enc_utf8) {
         # request transcoded result as UTF-8 octets!
-        $text = $self->_normalize($text, $self->{charset}, 0);
+        $text = _normalize($text, $self->{charset}, 0);
       }
       $self->{rendered_type} = $self->{type};
       $self->{rendered} = $self->{'visible_rendered'} = $text;
@@ -762,7 +852,7 @@ sub delete_header {
 
 # decode a header appropriately.  don't bother adding it to the pod documents.
 sub __decode_header {
-  my ( $self, $encoding, $cte, $data ) = @_;
+  my ( $encoding, $cte, $data ) = @_;
 
   if ( $cte eq 'B' ) {
     # base 64 encoded
@@ -780,13 +870,13 @@ sub __decode_header {
     # not possible since the input has already been limited to 'B' and 'Q'
     die "message: unknown encoding type '$cte' in RFC2047 header";
   }
-  return $self->_normalize($data, $encoding, 0);  # transcode to UTF-8 octets
+  return _normalize($data, $encoding, 0);  # transcode to UTF-8 octets
 }
 
 # Decode base64 and quoted-printable in headers according to RFC2047.
 #
 sub _decode_header {
-  my($self, $header_field_body, $header_field_name) = @_;
+  my($header_field_body, $header_field_name) = @_;
 
   return '' unless defined $header_field_body && $header_field_body ne '';
 
@@ -795,10 +885,10 @@ sub _decode_header {
   $header_field_body =~ s/\015?\012//gs;
 
   if ($header_field_name =~
-       /^ (?: (?: Received | (?:Resent-)? (?: Message-ID | Date ) |
-                  MIME-Version | References | In-Reply-To ) \z
-            | (?: List- | Content- ) ) /xsi ) {
+       /^ (?: Received | (?:Resent-)? (?: Message-ID | Date ) |
+              MIME-Version | References | In-Reply-To | List-.* ) \z /xsi ) {
     # Bug 6945: some header fields must not be processed for MIME encoding
+    # Bug 7466: leave out the Content-*
 
   } else {
     local($1,$2,$3);
@@ -816,7 +906,7 @@ sub _decode_header {
     # or plain US-ASCII
     $header_field_body =~
       s{ (?: = \? ([A-Za-z0-9_-]+) \? ([bqBQ]) \? ([^?]*) \? = ) }
-       { $self->__decode_header($1, uc($2), $3) }xsge;
+       { __decode_header($1, uc($2), $3) }xsge;
   }
 
 # dbg("message: _decode_header %s: %s", $header_field_name, $header_field_body);
@@ -918,7 +1008,7 @@ sub get_all_headers {
 	$self->{'truncated_header'} = 1;
 	last HEADER;
       }
-      $lines[$position] = $self->{header_order}->[$position].":".$contents;
+      $lines[$position] = $self->{header_order}->[$position].": ".$contents;
     }
   }
 
